

#include <Application.h>
#include <Window.h>
#include <View.h>
#include <Dragger.h>
#include <Alert.h>
#include <string.h>
#include <Screen.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>



#define P_IMPEXP
#include "Preferences.h"


#define INSTANTIATE_RETURN_TYPE GaugeView

#pragma export on

#define N_GAUGES 5
#define SEGMENT_WIDTH 128

class GaugeView :	/*	make it a replicant!	*/
	public BView
{
public:

								GaugeView(
									BRect			area,
									const char *		name,
									uint32			resize,
									uint32			flags);
								GaugeView(
									BMessage *		message);
								~GaugeView();
		void						GetPreferredSize(
									float *			width,
									float *			height);
		void						ResizeToPreferred();
		void						AttachedToWindow();

static	INSTANTIATE_RETURN_TYPE *					Instantiate(
									BMessage *		data);
		status_t					Archive(
									BMessage *		data,
									bool				deep = true) const;
		void						DrawSmall();
		void						Draw(
									BRect			area);
		void						Pulse();
		void						MessageReceived(
									BMessage *		message);

protected:

		float						fTextWidth;
		float						fTextHeight;
		float						fBaseLine;
		int						fGaugeValue[N_GAUGES];
		int						fGaugeMax[N_GAUGES];
		int						fChangeTicks[N_GAUGES];
		char **					fGaugeNames;
		bigtime_t					fLastUpdate;

		void						DoInit();
		void						SetGaugeColor(
									int				ix);
		rgb_color					GetGaugeColor(
									int ix);
private:

	typedef BView inherited;
};

#pragma export reset


Preferences gPrefsGroup("x-h+-gauges");
PreferenceSet gPrefs(gPrefsGroup, "window", true);

class GaugeWindow :
	public BWindow
{
public:
								GaugeWindow(
									BRect			area,
									const char *		name,
									window_type		kind,
									uint32			flags) :
									BWindow(area, name, kind, flags)
								{
									area.OffsetTo(B_ORIGIN);
									GaugeView *v = new GaugeView(area, name, B_FOLLOW_NONE, B_WILL_DRAW|B_PULSE_NEEDED);
									v->ResizeToPreferred();
									ResizeTo(v->Bounds().Width(), v->Bounds().Height()-1);
									AddChild(v);
								}
								~GaugeWindow()
								{
									BRect f(Frame());
									gPrefs.SetData("position", &f, sizeof(BRect), B_RECT_TYPE);
									gPrefs.Save();
								}

		bool						QuitRequested()
								{
									be_app->PostMessage(B_QUIT_REQUESTED);
									return true;
								}
};



int
main(
	int argc,
	char * argv[])
{
	BApplication app("application/x-h+-gauges");

	if (argc != 1)
	{
		if (strcmp(argv[1], "--install") || argc != 2)
		{
			fprintf(stderr, "usage: gauges [--install]\n");
			exit(1);
		}
		/* so it's an install */
		BView * v = new GaugeView(BRect(0,0,2*N_GAUGES-1,15), "Gauges", B_FOLLOW_NONE, B_WILL_DRAW|B_PULSE_NEEDED);
		BMessage msg(B_ARCHIVED_OBJECT);
		v->Archive(&msg);
		BMessenger messenger("application/x-vnd.Be-TSKB", -1, NULL);
		BMessage reply;
		status_t err = messenger.SendMessage(&msg, &reply) < B_OK ? 2 : 0;
		if (err == B_OK)
			reply.FindInt32("error", &err);
		if (err != B_OK) {
			fprintf(stderr, "install err: [%x] %s\n", err, strerror(err));
			reply.PrintToStream();
		}
		return err ? 2 : 0;
	}

	BRect wRect(10,30, 210,130);
	const void * data = NULL; ssize_t size = 0; uint32 type = B_RECT_TYPE;
	status_t res = gPrefs.GetData("position", data, size, type);
	if (!res)
	{
		memcpy(&wRect, data, size > sizeof(wRect) ? sizeof(wRect) : size);
		/*	fit rectangle on screen	*/
		BScreen sc;
		BRect fr = sc.Frame();
		if (wRect.left > fr.right-10)
			wRect.OffsetTo(fr.right-10, wRect.top);
		if (wRect.top > fr.bottom-10)
			wRect.OffsetTo(wRect.left, fr.bottom-10);
		if (wRect.right < fr.left+10)
			wRect.OffsetBy(fr.left-wRect.right+10, 0);
		if (wRect.bottom < fr.top+10)
			wRect.OffsetBy(0, fr.top-wRect.bottom+10);
	}
	GaugeWindow *w = new GaugeWindow(wRect, "Gauges", B_TITLED_WINDOW,
		B_NOT_RESIZABLE | B_NOT_ZOOMABLE);
	w->Show();
	app.Run();

	return 0;
}


static char * sGaugeNames[] =
{
	"PhysMem",
	"Semaphores",
	"Ports",
	"Threads",
	"Teams",
	NULL
};



GaugeView::GaugeView(
	BRect			area,
	const char *		name,
	uint32			resize,
	uint32			flags) :
	BView(area, name, resize, flags)
{
	DoInit();
}


GaugeView::GaugeView(
	BMessage *		message) :
	BView(message)
{
	DoInit();
}


void
GaugeView::DoInit()
{
	font_height height;
	be_plain_font->GetHeight(&height);
	fTextHeight = height.ascent+height.descent+height.leading;
	fBaseLine = height.ascent+height.leading;
	fGaugeNames = sGaugeNames;
	memset(fGaugeValue, 0, sizeof(int)*N_GAUGES);
	memset(fGaugeMax, 0, sizeof(int)*N_GAUGES);
	memset(fChangeTicks, 0, sizeof(int)*N_GAUGES);
	float maxWid = 0;
	for (int ix=0; ix<N_GAUGES; ix++)
	{
		float w = be_plain_font->StringWidth(fGaugeNames[ix]);
		if (w > maxWid)
			maxWid = w;
	}
	fTextWidth = maxWid;
	SetViewColor(B_TRANSPARENT_32_BIT);
	fLastUpdate = 0;
}


GaugeView::~GaugeView()
{
}


INSTANTIATE_RETURN_TYPE *
GaugeView::Instantiate(
	BMessage *		data)
{
	if (validate_instantiation(data, "GaugeView"))
		return new GaugeView(data);
	return NULL;
}


void
GaugeView::AttachedToWindow()
{
	inherited::AttachedToWindow();
	SetFlags(Flags() | B_PULSE_NEEDED);
	if (Window()->PulseRate() > 200000)
		Window()->SetPulseRate(200000);
	if (!CountChildren() && (Bounds().right > 15))
	{
		BRect r = Bounds();
		r.right = r.left+7;
		r.top = r.bottom-7;
		BDragger *dragger = new BDragger(r, this, B_FOLLOW_LEFT | B_FOLLOW_BOTTOM);
		if (!CountChildren())
			AddChild(dragger);
	}
}


status_t
GaugeView::Archive(
	BMessage *		data,
	bool				deep) const
{
	status_t code = inherited::Archive(data, deep);
	if (code == B_OK)
		code = data->AddString("class", "GaugeView");
	if (code == B_OK)
		code = data->AddString("add_on", "application/x-h+-gauges");
	return code;
}


rgb_color colors[] = {
	{255, 255, 255},
	{220, 220, 220},
	{120, 120, 120},
	{60, 60, 60},
	{0, 0, 0},
	{60, 180, 0},		/*	lots left	*/
	{150, 50, 50},	/*	red bar	*/
	{255, 128, 128},	/*	just changed	*/
	{255, 128, 0}		/*	at limit	*/
};

#define DECAY 32

rgb_color
GaugeView::GetGaugeColor(
	int ix)
{
	int delta = fChangeTicks[ix];
	int red = 128, green = 128, blue = 128;
	if (fGaugeMax[ix])
		red = (colors[8].red*fGaugeValue[ix]+colors[5].red*(fGaugeMax[ix]-fGaugeValue[ix]))/fGaugeMax[ix];
	if (fGaugeMax[ix])
		green = (colors[8].green*fGaugeValue[ix]+colors[5].green*(fGaugeMax[ix]-fGaugeValue[ix]))/fGaugeMax[ix];
	if (fGaugeMax[ix])
		blue = (colors[8].blue*fGaugeValue[ix]+colors[5].blue*(fGaugeMax[ix]-fGaugeValue[ix]))/fGaugeMax[ix];
	if (red > 255) red = 255; if (red < 0) red = 0;
	if (green > 255) green = 255; if (green < 0) green = 0;
	if (blue > 255) blue = 255; if (blue < 0) blue = 0;
	rgb_color ret;
	ret.alpha = 255;
	if (delta >= DECAY)
	{
		ret.red = red;
		ret.green = green;
		ret.blue = blue;
	}
	else
	{
		ret.red = (red*delta+(int)colors[7].red*(DECAY-delta))/DECAY;
		ret.green = (green*delta+(int)colors[7].green*(DECAY-delta))/DECAY;
		ret.blue = (blue*delta+(int)colors[7].blue*(DECAY-delta))/DECAY;
	}
	return ret;
}

void
GaugeView::SetGaugeColor(
	int				ix)
{
	SetHighColor(GetGaugeColor(ix));
}

void
GaugeView::DrawSmall()
{
	BRect r = Bounds();
	float bot = Bounds().bottom;
	SetHighColor(colors[3]);
	int offs = int(floor((r.Width()-2*N_GAUGES+1)/2));
	FillRect(r, B_SOLID_HIGH);
	BeginLineArray(N_GAUGES*2+1);
	for (int ix=0; ix<N_GAUGES; ix++)
	{
		if ((fGaugeMax[ix] > 0) && (fGaugeValue[ix] > 0))
		{
			AddLine(BPoint(ix*2+offs,bot), BPoint(ix*2+offs,bot*(fGaugeMax[ix]-fGaugeValue[ix])/fGaugeMax[ix]),
				GetGaugeColor(ix));
			AddLine(BPoint(ix*2+offs+1,bot), BPoint(ix*2+1+offs,bot*(fGaugeMax[ix]-fGaugeValue[ix])/fGaugeMax[ix]),
				GetGaugeColor(ix));
		}
		else
		{
			AddLine(BPoint(ix*2+offs,bot), BPoint(ix*2+offs,bot), GetGaugeColor(ix));
			AddLine(BPoint(ix*2+1+offs,bot), BPoint(ix*2+1+offs,bot), GetGaugeColor(ix));
		}
	}
	/* resource limit line */
	AddLine(BPoint(0,floor(bot*.2)), BPoint(Bounds().right,floor(bot*.2)), colors[6]);
	EndLineArray();
}


void
GaugeView::Draw(
	BRect			/* area */)
{
	if (Bounds().Width() < 16)
	{
		DrawSmall();
		return;
	}
	BRect r = Bounds();
	BeginLineArray(4);
	AddLine(r.RightTop(), r.RightBottom(), colors[2]);
	AddLine(r.LeftBottom(), r.RightBottom(), colors[2]);
	AddLine(r.LeftTop(), r.RightTop(), colors[0]);
	AddLine(r.LeftTop(), r.LeftBottom(), colors[0]);
	EndLineArray();
	for (int ix=0; ix<N_GAUGES; ix++)
	{
		float f = 1+ix*(fTextHeight+2);
		r.Set(1,f, fTextWidth+9, f+fTextHeight+1);
		SetLowColor(colors[1]);
		SetHighColor(colors[4]);
		FillRect(r, B_SOLID_LOW);
		DrawString(fGaugeNames[ix], BPoint(5, f+fBaseLine+1));
		r.left = r.right+1;
		r.right = r.right+SEGMENT_WIDTH;
		BeginLineArray(4);
		AddLine(r.LeftTop(), r.RightTop(), colors[2]);
		AddLine(r.LeftTop(), r.LeftBottom(), colors[2]);
		AddLine(r.RightTop(), r.RightBottom(), colors[0]);
		AddLine(r.LeftBottom(), r.RightBottom(), colors[0]);
		EndLineArray();
		r.InsetBy(1,1);
		float s = r.right;
		if ((fGaugeMax[ix] > 0) && (fGaugeValue[ix] > 0))
			r.right = r.left + (r.right-r.left)*fGaugeValue[ix]/fGaugeMax[ix];
		else
			r.right = r.left;
//		SetHighColor(colors[5]);
		SetGaugeColor(ix);
		FillRect(r);
		r.left = r.right+1;
		r.right = s;
		if (r.right >= r.left)
		{
			SetHighColor(colors[3]);
			FillRect(r);
		}
		SetHighColor(colors[6]);
		StrokeLine(BPoint(s-16, r.top), BPoint(s-16, r.bottom));
	}
	r.left = r.right+2;
	r.right = Bounds().right-1;
	r.top = Bounds().top+1;
	r.bottom = Bounds().bottom-1;
	FillRect(r, B_SOLID_LOW);
}


#define CHANGE(i,a) \
	if (fGaugeMax[i] != info.max_##a) { fChangeTicks[i] = -1; fGaugeMax[i] = info.max_##a; } \
	if (fGaugeValue[i] != info.used_##a) { fChangeTicks[i] = -1; fGaugeValue[i] = info.used_##a; } \
	fChangeTicks[i]++;

void
GaugeView::Pulse()
{
	inherited::Pulse();
	if (system_time() > fLastUpdate+10000000/DECAY)
	{
		system_info info;
		get_system_info(&info);
		CHANGE(0,pages);
		CHANGE(1,sems);
		CHANGE(2,ports);
		CHANGE(3,threads);
		CHANGE(4,teams);
		bool redraw = false;
		for (int ix=0; ix<5; ix++)
			if (fChangeTicks[ix]<=DECAY) redraw = true;
		if (redraw)
		{
			if (Bounds().Width() < 16)
			{
				Invalidate();
			}
			else
			{
				BRect r(fTextWidth+9,1,Bounds().right-1,Bounds().bottom-1);
				Invalidate(r);
			}
		}
		fLastUpdate = system_time();
	}
}


void
GaugeView::ResizeToPreferred()
{
	float wid = 200;
	float hei = 100;
	GetPreferredSize(&wid, &hei);
	ResizeTo(wid, hei);
}


void
GaugeView::GetPreferredSize(
	float *			width,
	float *			height)
{
	*width = fTextWidth + 12 + SEGMENT_WIDTH;
	*height = (fTextHeight+2) * N_GAUGES + 2;
}


void
GaugeView::MessageReceived(
	BMessage *		message)
{
	if (message->what == B_ABOUT_REQUESTED)
	{
		BAlert *alert = new BAlert("About",
			"Gauges displays the system's usage of various limited system resources.\n\n"
			"When a gauge reaches the red line, it's time to open your computer's cover and pour in more "
			"of the system resource listed for that gauge. System resources are found at better supermarkets "
			"everywhere; usually in the home electronics isle.\n\n"
			"Version 1.2" __DATE__ "\n"
			"Copyright © 1997-8 Jon Watte\n"
			"<gauges@mindcontrol.org>\n",
			"Duh!");
		alert->Go();
	}
	else
		inherited::MessageReceived(message);
}
